# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 1999-2007 Peter Thoeny, peter@thoeny.org
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# This module is based/inspired on Catalyst framework. Refer to
#
# http://search.cpan.org/~mramberg/Catalyst-Runtime-5.7010/lib/Catalyst.pm
# 
# for credits and liscence details.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

=begin twiki

---+!! package TWiki::Engine::Apache

Base class to all mod_perl engines.

Refer to TWiki::Engine documentation for explanation about methos below.

=cut

package TWiki::Engine::Apache;

use strict;
use base 'TWiki::Engine';
use URI;
use URI::http;
use URI::https;
use HTTP::Body;
use TWiki::Request::Upload;

use constant CHUNKSIZE => 4096;

BEGIN {
    # suppress stupid warning in CGI::Cookie
    if ( exists $ENV{MOD_PERL} ) {
        if ( !defined( $ENV{MOD_PERL_API_VERSION} )) {
            $ENV{MOD_PERL_API_VERSION} = 1;
        }
    }
}

sub prepareRequest {
    my ( $this, $req, $r ) = @_;
    $this->{r} = $r;
}

# SMELL: Doesn't support remotePort. Anyway, is this really needed?
sub prepareConnection {
    my ( $this, $req ) = @_;
    $req->method( $this->{r}->method );
    $req->remoteAddress( $this->{r}->connection->remote_ip );
    $req->remoteUser( $this->{r}->user );
    if ($INC{'Apache2/ModSSL.pm'}) {
        $req->secure ($this->{r}->connection->is_https ? 1 : 0);
    }
    else {
        my $https = $this->{r}->subprocess_env('HTTPS');
        $req->secure( defined $https && uc($https) eq 'ON' ? 1 : 0);
    }
}

sub prepareQueryParameters {
    my ( $this, $req ) = @_;
    $req->uri( $this->{r}->uri );
    my $query = $this->{r}->args;
    $this->SUPER::prepareQueryParameters($req, $query )
        if $query;
}

sub prepareHeaders {
    my ( $this, $req ) = @_;
    $req->header( %{ $this->{r}->headers_in() } );
}

sub preparePath {
    my ( $this, $req ) = @_;
    my $location = $this->{r}->location();
    ( my $pathInfo = $this->{r}->uri() ) =~ s/^$location//;
    $req->pathInfo( $pathInfo );
}

sub prepareBody {
    my ( $this, $req ) = @_;
    my $contentLength = $req->header('Content-Length') || 0;
    return unless $contentLength > 0;

    $this->{body} = new HTTP::Body( $req->header('Content-Type'), $contentLength );
    my $bytesRead = 0;
READ:
    while ( $bytesRead < $contentLength ) {
        my $remaining = $contentLength - $bytesRead;
        my $chunk     = $remaining > CHUNKSIZE ? CHUNKSIZE : $remaining;
        my $rc        = $this->{r}->read( my $buffer, $chunk );
        if ( defined $rc ) {
            $this->{body}->add($buffer);
            $bytesRead += $rc;
        }
        else {
            last READ;
        }
    }
    if ( $bytesRead != $contentLength ) {
        throw TWiki::EngineException( '400', 'Bad Request: wrong Content-Length header' );
    }
}

sub prepareBodyParameters {
    my ( $this, $req ) = @_;
    return unless $this->{body};
    while ( my ( $key, $value ) = each %{ $this->{body}->param } ) {
        $req->bodyParam( -name => $key, -value => $value );
    }
}

sub prepareUploads {
    my ( $this, $req ) = @_;
    return unless $this->{body};
    my %uploads;
    while ( my ( $key, $value ) = each %{ $this->{body}->upload } ) {
        $req->bodyParam(
            -name  => $value->{name},
            -value => $value->{filename},
        );
        $uploads{ $value->{name} } = new TWiki::Request::Upload(
            name    => $value->{name},
            headers => $value->{headers},
            tmpname => $value->{tempname},
        );
    }
    delete $this->{body};
    $req->uploads( \%uploads );
}

sub finalizeUploads {
    my ( $this, $res, $req ) = @_;
    $req->delete($_) foreach keys %{ $req->uploads };
}

sub finalizeHeaders { 
    my ( $this, $res, $req ) = @_;
    $this->SUPER::finalizeHeaders( $res, $req );
    $this->finalizeCookies($res);
    my ( $status ) = ( $res->status || "200" ) =~ /^(\d\d\d)/o;
    
    $this->{r}->status( $res->status || 200 );
    while ( my ( $header, $value ) = each %{ $res->headers } ) {
        foreach ( ref($value) eq 'ARRAY' ? @$value : ($value) ) {
            unless ( $header eq 'Set-Cookie' ) {
                $this->{r}->headers_out->add( $header => $_ );
            }
            else { # Send cookies even if an error occour
                $this->{r}->err_headers_out->add( $header => $_ )
            }
        }
    }
    $this->{r}->content_type( $res->getHeader('Content-Type') || 'text/plain' );
}

sub write {
    my ($this, $data) = @_;
    if ( !$this->{r}->connection->aborted && defined $data ) {
        return $this->{r}->print($data);
    }
    return undef;
}

1;
