# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 1999-2007 Peter Thoeny, peter@thoeny.org
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

=begin twiki

---+!! package TWiki::Engine::HTTP

Class that implements an HTTP stand alone server capable of
dealing with twiki and static requests.

Refer to TWiki::Engine documentation for explanation about methos below.

=cut

package TWiki::Engine::HTTP;

use strict;
use base 'TWiki::Engine';
BEGIN {
    require LWP::MediaTypes;
    LWP::MediaTypes::read_media_types($TWiki::cfg{MimeTypesFileName});
}
use HTTP::Daemon;
use Socket;
use HTTP::Request;
use HTTP::Headers;
use HTTP::Body;
use Errno 'EWOULDBLOCK';
use TWiki::Request;
use TWiki::Request::Upload;
use TWiki::Response;
use Cwd;

use constant CHUNKSIZE => 4096;
use constant READ_TIMEOUT => 60;

sub new {
    my $this = shift;
    $this = $this->SUPER::new(@_);
    $this->{dir} = getcwd;
    return $this;
}

sub run {
    my $this = shift;
    my %args = (
        port   => 3000,
        host   => 0,
        detach => 0,
        binUrl => $TWiki::cfg{ScriptUrlPath},
        pubUrl => $TWiki::cfg{PubUrlPath},
        @_
    );
    my $httpd = HTTP::Daemon->new(
        LocalPort => $args{port},
        LocalHost => $args{host},
        Listen    => 50,
        Reuse     => 1
    ) or die "Could not setup server: $!";
    my $binUrl = $args{binUrl};
    my $pubUrl = $args{pubUrl};
    if ( $args{detach} ) {
        detach();
    }
    else {
        my $ip = $this->_canonicalIP($args{host});
        $ip = '127.0.0.1' if $ip eq '0.0.0.0';
        print "TWiki HTTP Engine started. Visit me at ";
        print 'http://' . $ip;
        print ':' . $args{port} if $args{port} != 80;
        print "$args{binUrl}/view$TWiki::cfg{ScriptSuffix}\n";
    }

    local $SIG{PIPE} = 'IGNORE';

  CLIENT:
    while ( my $client = $httpd->accept ) {
        my $r = $client->get_request('headers_only_please');
        next CLIENT unless $r;
        if ( $r->method eq 'RESTART' ) {
            my $peerIp = $this->_canonicalIP($client->peerhost);
            my $ip     = $this->_canonicalIP($httpd->sockhost);
            $ip = '127.0.0.1' if $ip eq '0.0.0.0';
            last CLIENT if $peerIp eq '127.0.0.1' || $peerIp eq $ip;
        }
        elsif ( $r->method eq 'GET' && $r->uri->path =~ m#^$pubUrl#o ) {
            my $pubDir = $TWiki::cfg{PubDir};
            my $file = $r->uri->path;
            $file =~ s/\.\.//g;
            $file =~ s#^$pubUrl#$pubDir#o;
            $client->send_file_response($file) unless $file =~ /\.ht.+/;
        }
        elsif ( $r->uri->path =~ m#^$binUrl#o ) {
            TWiki::UI::handleRequest( $r, $client, $binUrl );
        }
        else {
            $client->send_error(404); # Not found
        }
        $client->close;
        undef $client;
    }
    $httpd->close;

    local $SIG{CHLD} = 'DEFAULT';
    wait unless $args{detach};
    ### if the standalone server was invoked with perl -I .. we will loose
    ### those include dirs upon re-exec. So add them to PERL5LIB, so they
    ### are available again for the exec'ed process --kane
    use Config;
    $ENV{PERL5LIB} .= join $Config{path_sep}, @INC;
    $ENV{PATH} = $TWiki::cfg{SafeEnvPath};
    my $perl = $Config{perlpath};
    if ( $^O ne 'VMS' ) {
        $perl .= $Config{_exe}
            unless $perl =~ m/$Config{_exe}$/i;
    }
    chdir TWiki::Sandbox::untaintUnchecked($this->{dir});
    my $script = TWiki::Sandbox::untaintUnchecked($0);
    my $args   = TWiki::Sandbox::untaintUnchecked(join( ' ', @ARGV ));
    exec $perl . ' -wT "'. $script . '" ' . $args;
}

=begin twiki

---++ StaticMethod detach()

Daemonize process. Currently not portable...

=cut

sub detach {
    fork && exit;
    require POSIX;
    umask(0);
    chdir '/';
    open STDIN,  "+</dev/null" or die $!;
    open STDOUT, ">&STDIN"     or die $!;
    open STDERR, ">&STDIN"     or die $!;
    POSIX::setsid();
}

sub _canonicalIP {
    return inet_ntoa(inet_aton($_[1]));
}

sub prepareRequest {
    my ( $this, $req, $r, $client, $binUrl ) = @_;
    $this->{r}      = $r;
    $this->{client} = $client;
    $this->{binUrl} = $binUrl;
}

sub prepareConnection { 
    my ( $this, $req ) = @_;
    $req->method( $this->{r}->method );
    $req->remoteAddress( $this->_canonicalIP($this->{client}->peerhost) );
    $req->remotePort( $this->{client}->peerport );
    $req->secure(0); # Doesn't support https yet
    # remoteUser is set in prepareHeaders, cause of authentication
}

sub prepareQueryParameters {
    my ( $this, $req ) = @_;
    $req->uri( $this->{r}->uri );
    my $query = $req->uri->query;
    $this->SUPER::prepareQueryParameters( $req, $query )
      if $query;
}

sub prepareHeaders {
    my ( $this, $req ) = @_;
    $req->header( $_ => $this->{r}->header($_) )
      foreach $this->{r}->header_field_names;
    if ( $TWiki::cfg{BehindProxy} ) {
        if ( my $source = $req->header('X-Forwarded-For') ) {
            my $ip = (split /[, ]+/, $source)[-1];
            $req->remoteAddress($1) 
                if defined $ip and $ip =~ /^((?:\d{1,3}\.){3}\d{1,3})$/;
        }
    }
    my $user;
    if ( $TWiki::cfg{LoginManager} eq 'TWiki::LoginManager::ApacheLogin' ) {
        my ( $pass, $htPasswd );
        ( $user, $pass ) = $this->{r}->authorization_basic();
        if ( defined $user ) {
            require TWiki::Users::HtPasswdUser;
            $htPasswd = new TWiki::Users::HtPasswdUser();
            $user = undef unless $htPasswd->checkPassword( $user, $pass );
        }
    }
    $req->remoteUser( $user );
}

sub preparePath { 
    my ( $this, $req ) = @_;
    my $uri = $req->uri;
    my $binUrl = $this->{binUrl};
    ( my $pathInfo = $req->uri->path ) =~ s#^.*$binUrl##;
    $req->pathInfo( $pathInfo );
    if ( $TWiki::cfg{LoginManager} eq 'TWiki::LoginManager::ApacheLogin' &&
         ! defined $req->remoteUser ) 
    {
        foreach my $authScript ( split(/[,\s]+/, $TWiki::cfg{AuthScripts}) ) {
            if ( $req->action eq $authScript ) {
                my $res = new TWiki::Response();
                $res->header( -type => 'text/html', -status => 401 );
                $res->pushHeader( 'WWW-Authenticate' => "Basic realm=$TWiki::cfg{AuthRealm}" );
                my $html = CGI::start_html('401 Authorization required');
                $html .= CGI::h1('Authentication Error') . CGI::end_html();
                $res->body($html);
                throw TWiki::EngineException(401, 'Authentication Error', $res);
            }
        }
    }
}

sub prepareBody {
    my ( $this, $req ) = @_;
    my $contentLength = $req->header('Content-Length') || 0;
    return unless $contentLength > 0;

    $this->{body} =
      new HTTP::Body( $req->header('Content-Type'), $contentLength );
    my $bytesRead = 0;
    if ( ${ $this->{client} }->{httpd_rbuf} ) {
        use bytes;
        $this->{body}->add( ${ $this->{client} }->{httpd_rbuf} );
        $bytesRead = length( ${ $this->{client} }->{httpd_rbuf} );
        ${ $this->{client} }->{httpd_rbuf} = '';
    }

    $this->{client}->blocking(0);
    my $rin = "";
    vec ( $rin, $this->{client}->fileno, 1 ) = 1;

READ:
    while ( $bytesRead < $contentLength
        && select( $rin, undef, undef, READ_TIMEOUT ) )
    {
        my $remaining = $contentLength - $bytesRead;
        my $chunk     = $remaining > CHUNKSIZE ? CHUNKSIZE : $remaining;
        my $rc        = sysread( $this->{client}, my $buffer, $chunk );
        if ( defined $rc ) {
            $this->{body}->add($buffer);
            $bytesRead += $rc;
        }
        elsif ( $! == EWOULDBLOCK ) {
            next READ;
        }
        else {
            last READ;
        }
    }
    if ( $bytesRead != $contentLength ) {
        throw TWiki::EngineException( '400',
            'Bad Request: wrong Content-Length header' );
    }
    $this->{client}->blocking(1);
}

sub prepareBodyParameters {
    my ( $this, $req ) = @_;
    return unless $this->{body};
    while ( my ( $key, $value ) = each %{ $this->{body}->param } ) {
        $req->bodyParam( -name => $key, -value => $value );
    }
}

sub prepareUploads {
    my ( $this, $req ) = @_;
    return unless $this->{body};
    my %uploads;
    while ( my ( $key, $value ) = each %{ $this->{body}->upload } ) {
        $req->bodyParam(
            -name  => $value->{name},
            -value => $value->{filename},
        );
        $uploads{ $value->{name} } = new TWiki::Request::Upload(
            name    => $value->{name},
            headers => $value->{headers},
            tmpname => $value->{tempname},
        );
    }
    delete $this->{body};
    $req->uploads( \%uploads );
}

sub finalizeUploads {
    my ( $this, $res, $req ) = @_;
    $req->delete($_) foreach keys %{ $req->uploads };
}

sub finalizeHeaders { 
    my ( $this, $res, $req ) = @_;
    $this->SUPER::finalizeHeaders( $res, $req );
    $this->finalizeCookies($res);
    my ( $status ) = ( $res->status || "200" ) =~ /^(\d\d\d)/o;
    my $response = HTTP::Response->new($status);
    $response->protocol( $this->{r}->protocol );
    $response->header( $_ => $res->getHeader($_) )
        foreach $res->getHeader();
    $response->header( Connection => 'close' ); # No keep-alive yet
    my $headerStr = $response->as_string($this->CRLF);
    if ( $res->body ) {
        # Concats header and body to reduce number of TCP packets...
        $this->{outBuffer} = $headerStr;
    }
    else {
        $this->{client}->syswrite($headerStr);
    }
}

sub write {
    my ($this, $data) = @_;
    if ( $this->{outBuffer} ) {
        $data = $this->{outBuffer} . $data;
        delete $this->{outBuffer};
    }
    $this->{client}->syswrite($data);
}

1;
