# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 1999-2007 Peter Thoeny, peter@thoeny.org
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# Part of this module is based on "bin/twiki" script, which is:
#    Copyright (C) 2005 Martin at Cleaver.org
#    Copyright (C) 2005-2007 TWiki Contributors
# 
# and also based/inspired on Catalyst framework, whose Author is
# Sebastian Riedel. Refer to
#
# http://search.cpan.org/~mramberg/Catalyst-Runtime-5.7010/lib/Catalyst.pm
# 
# for more credit and liscence details.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

=pod

---+!! package TWiki::UI

Coordinator of execution flow and service functions used by the UI packages

=cut

package TWiki::UI;
use strict;

BEGIN {
    if ( $ENV{MOD_PERL} ) {
        my ( $software, $version ) =
            $ENV{MOD_PERL} =~ m{^(\S+)/(\d+(?:[\.\_]\d+)+)};

        $version =~ s/_//g;
        $version =~ s/(\.[^.]+)\./$1/g;

        if ( $software eq 'mod_perl' ) {
            if ( !$TWiki::cfg{Engine} ) {
                if ( $version >= 1.99922 ) {
                    $TWiki::cfg{Engine} = 'TWiki::Engine::Apache2::MP20';
                }
                elsif ( $version >= 1.9901 ) {
                    $TWiki::cfg{Engine} = 'TWiki::Engine::Apache2::MP19';
                }
                elsif ( $version >= 1.24 ) {
                    $TWiki::cfg{Engine} = 'TWiki::Engine::Apache::MP13';
                }
                else {
                    die qq(Unsupported mod_perl version: $ENV{MOD_PERL} );
                }
            }
            *handler = \&handleRequest;
        }
    }

    $TWiki::cfg{SwitchBoard} = {
        'attach'      => [ 'TWiki::UI::Upload',     'attach',        { }              ],
        'changes'     => [ 'TWiki::UI::Changes',    'changes',       { changes => 1 } ],
        'edit'        => [ 'TWiki::UI::Edit',       'edit',          { edit    => 1 } ],
        'login'       => [ undef,                   'login',         { }              ],
        'logon'       => [ undef,                   'logon',         { }              ],
        'manage'      => [ 'TWiki::UI::Manage',     'manage',        { }              ],
        'oops'        => [ 'TWiki::UI::Oops',       'oops_cgi',      { oops    => 1 } ],
        'preview'     => [ 'TWiki::UI::Preview',    'preview',       { preview => 1 } ],
        'rdiffauth'   => [ 'TWiki::UI::RDiff',      'diff',          { diff    => 1 } ],
        'rdiff'       => [ 'TWiki::UI::RDiff',      'diff',          { diff    => 1 } ],
        'register'    => [ 'TWiki::UI::Register',   'register_cgi',  { }              ],
        'rename'      => [ 'TWiki::UI::Manage',     'rename',        { }              ],
        'resetpasswd' => [ 'TWiki::UI::Register',   'resetPassword', { }              ],
        'rest'        => [ 'TWiki::UI::Rest',       'rest',          { rest    => 1 } ],
        'save'        => [ 'TWiki::UI::Save',       'save',          { save    => 1 } ],
        'search'      => [ 'TWiki::UI::Search',     'search',        { search  => 1 } ],
        'statistics'  => [ 'TWiki::UI::Statistics', 'statistics',    { }              ],
        'upload'      => [ 'TWiki::UI::Upload',     'upload',        { }              ],
        'viewauth'    => [ 'TWiki::UI::View',       'view',          { view    => 1 } ],
        'viewfile'    => [ 'TWiki::UI::View',       'viewfile',      { }              ],
        'view'        => [ 'TWiki::UI::View',       'view',          { view    => 1 } ],
    };
}

use Error qw(:try);
use Assert;

use TWiki;
use TWiki::Request;
use TWiki::Response;
use TWiki::OopsException;
use TWiki::EngineException;
use CGI;

# Used to lazily load UI handler modules
our %isInitialized = ();

sub TRACE_PASSTHRU {
    # Change to a 1 to trace passthrough
    0;
};

=begin twiki

---++ StaticMethod handleRequest(...)

Main coordinator of request-process-response cycle.

... are any parameters passed by engine, which are forwarded to prepare

=cut

sub handleRequest {
    my $req = prepare(@_);
    if ( ref $req ) {
        my $res = dispatch($req);
        finalize( $res, $req );
        return $res->status;
    }
    return $req; # Error status code returned by prepare
}

=begin twiki

---++ StaticMethod prepare(...) -> $req

Initialize a TWiki::Request object by calling many engine methods
and returns it.

   * ... are any parameters to pass to engine's prepareRequest method

=cut

sub prepare {
    my @p = @_;
    my $req;
    try {
        $req = TWiki::Request->new();
        $TWiki::engine->prepareRequest( $req, @p );
        $TWiki::engine->prepareConnection($req);
        $TWiki::engine->prepareQueryParameters($req);
        $TWiki::engine->prepareHeaders($req);
        $TWiki::engine->preparePath($req);
        $TWiki::engine->prepareCookies($req);
        $TWiki::engine->prepareBody($req);
        $TWiki::engine->prepareBodyParameters($req);
        $TWiki::engine->prepareUploads($req);
        $TWiki::engine->prepareParameters($req);
    }
    catch TWiki::EngineException with {
        my $e   = shift;
        my $res = $e->{response};
        unless (defined $res) {
            $res = new TWiki::Response();
            $res->header( -type => 'text/html', -status => $e->{status} );
            my $html = CGI::start_html( $e->{status} . ' Bad Request' );
            $html .= CGI::h1('Bad Request');
            $html .= CGI::p( $e->{reason} );
            $html .= CGI::end_html();
            $res->body($html);
        }
        $TWiki::engine->finalizeError($res);
        return $e->{status};
    }
    otherwise {
        my $e = shift;
        my $res = new TWiki::Response;
        $res->header( -type => 'text/plain' );
        if (DEBUG) {
            # output the full message and stacktrace to the browser
            $res->body( $e->stringify() );
        }
        else {
            my $mess = $e->stringify();
            print STDERR $mess;
            # tell the browser where to look for more help
            my $text = 'TWiki detected an internal error - please check your TWiki logs and webserver logs for more information.'."\n\n";
            $mess =~ s/ at .*$//s;
            # cut out pathnames from public announcement
            $mess =~ s#/[\w./]+#path#g;
            $text .= $mess;
            $res->body($text);
        }
        $TWiki::engine->finalizeError( $res );
        return 500; # Internal server error
    };
    return $req;
}

=begin twiki

---++ StaticMethod dispath($req) -> $res

Based on $req->action pass control to correct UI handler
and returns the TWiki::Response object generated.
   * =$req= is the TWiki::Request object

=cut

sub dispatch {
    my $req = shift;

    my $res;
    my $dispatcher = $TWiki::cfg{SwitchBoard}{$req->action()};
    unless (defined $dispatcher && ref($dispatcher) eq 'ARRAY') {
        $res = new TWiki::Response();
        $res->header(-type => 'text/html', -status => '404');
        my $html = CGI::start_html('404 Not Found');
        $html .=   CGI::h1('Not Found');
        $html .=   CGI::p("The requested URL " . $req->uri . " was not found on this server.");
        $html .=   CGI::end_html();
        $res->body($html);
        return $res;
    }
    my ( $package, $function, $context ) = @$dispatcher;

    if ($package && !$isInitialized{$package}) {
        eval qq(use $package);
        die $@ if $@;
        $isInitialized{$package} = 1;
    }

    my $sub;
    $sub  = $package.'::' if $package;
    $sub .= $function;

    if ( UNIVERSAL::isa($TWiki::engine, 'TWiki::Engine::CLI') ) {
        $context->{command_line} = 1;
    }

    $res = execute($req, \&$sub, %$context );
    return $res;
}

=begin twiki

---++ StaticMethod execute($req, $sub, %initialContext) -> $res

Creates a TWiki session object with %initalContext and calls
$sub method. Returns the TWiki::Response object generated

=cut

sub execute {
    my ($req, $sub, %initialContext ) = @_;

    my $cache = $req->param('twiki_redirect_cache');
    # Never trust input data from a query. We will only accept an MD5 32 character string
    if ( $cache && $cache =~ /^([a-f0-9]{32})$/ ) {
        $cache = $1;
        # Read cached post parameters
        my $passthruFilename =
          $TWiki::cfg{WorkingDir} . '/tmp/passthru_' . $cache;
        if ( open( F, '<', $passthruFilename ) ) {
            local $/;
            if (TRACE_PASSTHRU) {
                print STDERR "Passthru: Loading cache for ", $req->url(),
                  '?', $req->query_string(), "\n";
                print STDERR <F>, "\n";
                close(F);
                open( F, '<' . $passthruFilename );
            }
            $req = new TWiki::Request( \*F, $req );
            close(F);
            unlink($passthruFilename);
            print STDERR "Passthru: Loaded and unlinked $passthruFilename\n"
              if TRACE_PASSTHRU;
        }
        else {
            print STDERR "Passthru: Could not find $passthruFilename\n"
              if TRACE_PASSTHRU;
        }
    }

    my $session = new TWiki( $req->remoteUser, $req, \%initialContext );
    my $res;

    try {
        $session->{users}->{loginManager}->checkAccess();
        &$sub( $session );
        $res = $session->{response};
    } 
    catch TWiki::AccessControlException with {
        my $e = shift;
        unless( $session->{users}->{loginManager}->forceAuthentication() ) {
            # Login manager did not want to authenticate, perhaps because
            # we are already authenticated.
            my $exception = new TWiki::OopsException(
                'accessdenied',
                web => $e->{web}, topic => $e->{topic},
                def => 'topic_access',
                params => [ $e->{mode}, $e->{reason} ] );

            $exception->redirect( $session );
        }
        $res = $session->{response};
    }
    catch TWiki::OopsException with {
        shift->redirect( $session );
        $res = $session->{response};
    }
    catch Error::Simple with {
        my $e = shift;
        $res = new TWiki::Response;
        $res->header( -type => 'text/plain' );
        if (DEBUG) {
            # output the full message and stacktrace to the browser
            $res->body( $e->stringify() );
        }
        else {
            my $mess = $e->stringify();
            print STDERR $mess;
            $session->writeWarning($mess);
            # tell the browser where to look for more help
            my $text = 'TWiki detected an internal error - please check your TWiki logs and webserver logs for more information.'."\n\n";
            $mess =~ s/ at .*$//s;
            # cut out pathnames from public announcement
            $mess =~ s#/[\w./]+#path#g;
            $text .= $mess;
            $res->body($text);
        }
    }
    otherwise {
        $res = new TWiki::Response;
        $res->header(-type => 'text/plain');
        $res->body( "Unspecified error" );
    };

    $session->finish();
    return $res;
}

=begin twiki

---++ StaticMethod finalize($res, $req)

Finalizes the request by calling many engine's methods to send response to
client and take any appropriate finalize actions, such as delete temporary
files.
   * =$res= is the TWiki::Response object
   * =$req= it the TWiki::Request object. Needed to finalizeUploads

=cut

sub finalize {
    my ( $res, $req ) = @_;
    $TWiki::engine->finalizeUploads($res, $req);
    $TWiki::engine->finalizeHeaders($res, $req);
    $TWiki::engine->finalizeBody($res);
}

=begin twiki

---++ StaticMethod login($session)

Handler to "login" action.
   * =$session= is a TWiki session object

=cut

sub login {
  my $session = shift;
  $session->{users}->{loginManager}->login( $session->{request}, $session );
}

=pod twiki

---++ StaticMethod checkWebExists( $session, $web, $topic, $op )

Check if the web exists. If it doesn't, will throw an oops exception.
 $op is the user operation being performed.

=cut

sub checkWebExists {
    my ( $session, $webName, $topic, $op ) = @_;
    ASSERT($session->isa( 'TWiki')) if DEBUG;

    unless ( $session->{store}->webExists( $webName ) ) {
        throw
          TWiki::OopsException( 'accessdenied',
                                def => 'no_such_web',
                                web => $webName,
                                topic => $topic,
                                params => [ $op ] );
    }
}

=pod

---++ StaticMethod topicExists( $session, $web, $topic, $op ) => boolean

Check if the given topic exists, throwing an OopsException
if it doesn't. $op is the user operation being performed.

=cut

sub checkTopicExists {
    my ( $session, $webName, $topic, $op ) = @_;
    ASSERT($session->isa( 'TWiki')) if DEBUG;

    unless( $session->{store}->topicExists( $webName, $topic )) {
        throw TWiki::OopsException( 'accessdenied',
                                    def => 'no_such_topic',
                                    web => $webName,
                                    topic => $topic,
                                    params => [ $op ] );
    }
}

=pod twiki

---++ StaticMethod checkMirror( $session, $web, $topic )

Checks if this web is a mirror web, throwing an OopsException
if it is.

=cut

sub checkMirror {
    my ( $session, $webName, $topic ) = @_;
    ASSERT($session->isa( 'TWiki')) if DEBUG;

    my( $mirrorSiteName, $mirrorViewURL ) =
      $session->readOnlyMirrorWeb( $webName );

    return unless ( $mirrorSiteName );

    throw Error::Simple(
        "This is a mirror site $mirrorSiteName, $mirrorViewURL" );
}

=pod twiki

---++ StaticMethod checkAccess( $web, $topic, $mode, $user )

Check if the given mode of access by the given user to the given
web.topic is permissible, throwing a TWiki::OopsException if not.

=cut

sub checkAccess {
    my ( $session, $web, $topic, $mode, $user ) = @_;
    ASSERT($session->isa( 'TWiki')) if DEBUG;

    unless( $session->security->checkAccessPermission(
        $mode, $user, undef, undef, $topic, $web )) {
        throw TWiki::OopsException( 'accessdenied',
                                    def => 'topic_access',
                                    web => $web,
                                    topic => $topic,
                                    params =>
                                      [ $mode,
                                        $session->security->getReason()]);
    }
}

=pod

---++ StaticMethod readTemplateTopic( $session, $theTopicName ) -> ( $meta, $text )

Read a topic from the TWiki web, or if that fails from the current
web.

=cut

sub readTemplateTopic {
    my( $session, $theTopicName ) = @_;
    ASSERT($session->isa( 'TWiki')) if DEBUG;

    $theTopicName =~ s/$TWiki::cfg{NameFilter}//go;

    my $web = $TWiki::cfg{SystemWebName};
    if( $session->{store}->topicExists( $session->{webName}, $theTopicName )) {
        # try to read from current web, if found
        $web = $session->{webName};
    }
    return $session->{store}->readTopic(
        $session->{user}, $web, $theTopicName, undef );
}

1;
